# Copyright (c) 2025, Huawei Technologies.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import json
from pathlib import Path
from werkzeug import wrappers, Response, exceptions
from tensorboard.backend import http_util
from ..service.graph_service import GraphService
from ..utils.graph_utils import GraphUtils


class GraphView:

    # 静态文件路由
    @staticmethod
    @wrappers.Request.application
    def static_file_route(request):
        filename = os.path.basename(request.path)
        extension = os.path.splitext(filename)[1]
        if extension == '.html':
            mimetype = 'text/html'
        elif extension == '.js':
            mimetype = 'application/javascript'
        else:
            mimetype = 'application/octet-stream'
        current_dir = Path(__file__).resolve().parent
        server_dir = current_dir.parent.parent
        filepath = server_dir / "static" / filename
        try:
            with open(filepath, 'rb') as infile:
                contents = infile.read()
        except IOError as e:
            raise exceptions.NotFound('404 Not Found') from e
        return Response(contents, content_type=mimetype, headers={"X-Content-Type-Options": "nosniff"})

    @staticmethod
    @wrappers.Request.application
    def load_meta_dir(request):
        """Scan logdir for directories containing .vis files, modified to return a tuple of (run, tag)."""
        result = GraphService.load_meta_dir()
        response = http_util.Respond(request, result, "application/json")
        return response

    # 读取当前图数据
    @staticmethod
    @wrappers.Request.application
    def load_graph_data(request):
        run = request.args.get("run")
        tag = request.args.get("tag")
        return Response(
            GraphService.load_graph_data(run, tag),
            mimetype='text/event-stream',
            headers={
                'Cache-Control': 'no-cache',
                'Connection': 'close',  # TCP链接不复用，请求结束释放资源
                "X-Content-Type-Options": "nosniff",
            }
        )

    # 获取当前图数据配置信息
    @staticmethod
    @wrappers.Request.application
    def load_graph_config_info(request):
        run = request.args.get("run")
        tag = request.args.get("tag")
        result = GraphService.load_graph_config_info(run, tag)
        # 创建响应对象
        response = http_util.Respond(request, result, "application/json")
        return response

    # 获取当前图所有节点列表
    @staticmethod
    @wrappers.Request.application
    def load_graph_all_node_list(request):
        run = request.args.get("run")
        tag = request.args.get("tag")
        micro_step = request.args.get('microStep')
        result = GraphService.load_graph_all_node_list(run, tag, micro_step)
        response = http_util.Respond(request, result, "application/json")
        return response

    # 展开关闭节点
    @staticmethod
    @wrappers.Request.application
    def change_node_expand_state(request):
        node_info = GraphUtils.safe_json_loads(request.args.get("nodeInfo"))
        meta_data = GraphUtils.safe_json_loads(request.args.get("metaData"))
        hierarchy = GraphService.change_node_expand_state(node_info, meta_data)
        return http_util.Respond(request, json.dumps(hierarchy), "application/json")

    # 更新当前图节点信息
    @staticmethod
    @wrappers.Request.application
    def update_hierarchy_data(request):
        graph_type = request.args.get("graphType")
        hierarchy = GraphService.update_hierarchy_data(graph_type)
        return http_util.Respond(request, json.dumps(hierarchy), "application/json")

    # 获取当前节点对应节点的信息看板数据
    @staticmethod
    @wrappers.Request.application
    def get_node_info(request):
        node_info = GraphUtils.safe_json_loads(request.args.get("nodeInfo"))
        meta_data = GraphUtils.safe_json_loads(request.args.get("metaData"))
        node_detail = GraphService.get_node_info(node_info, meta_data)
        return http_util.Respond(request, json.dumps(node_detail), "application/json")

    # 根据配置文件添加匹配节点
    @staticmethod
    @wrappers.Request.application
    def add_match_nodes_by_config(request):
        config_file = request.args.get("configFile")
        meta_data = GraphUtils.safe_json_loads(request.args.get("metaData"))
        match_result = GraphService.add_match_nodes_by_config(config_file, meta_data)
        return http_util.Respond(request, json.dumps(match_result), "application/json")

    # 添加匹配节点
    @staticmethod
    @wrappers.Request.application
    def add_match_nodes(request):
        npu_node_name = request.args.get("npuNodeName")
        bench_node_name = request.args.get("benchNodeName")
        meta_data = GraphUtils.safe_json_loads(request.args.get("metaData"))
        match_result = GraphService.add_match_nodes(npu_node_name, bench_node_name, meta_data)
        return http_util.Respond(request, json.dumps(match_result), "application/json")

    # 取消节点匹配
    @staticmethod
    @wrappers.Request.application
    def delete_match_nodes(request):
        npu_node_name = request.args.get("npuNodeName")
        bench_node_name = request.args.get("benchNodeName")
        meta_data = GraphUtils.safe_json_loads(request.args.get("metaData"))
        match_result = GraphService.delete_match_nodes(npu_node_name, bench_node_name, meta_data)
        return http_util.Respond(request, json.dumps(match_result), "application/json")

    # 保存匹配节点列表
    @staticmethod
    @wrappers.Request.application
    def save_data(request):
        meta_data = GraphUtils.safe_json_loads(request.args.get("metaData"))
        save_result = GraphService.save_data(meta_data)
        return http_util.Respond(request, json.dumps(save_result), "application/json")

    # 更新颜色信息
    @staticmethod
    @wrappers.Request.application
    def update_colors(request):
        run = request.args.get('run')
        colors = GraphUtils.safe_json_loads(request.args.get('colors'))
        update_result = GraphService.update_colors(run, colors)
        return http_util.Respond(request, json.dumps(update_result), "application/json")

    # 保存匹配关系
    @staticmethod
    @wrappers.Request.application
    def save_matched_relations(request):
        meta_data = GraphUtils.safe_json_loads(request.args.get("metaData"))
        save_result = GraphService.save_matched_relations(meta_data)
        return http_util.Respond(request, json.dumps(save_result), "application/json")



# Copyright (c) 2025, Huawei Technologies.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import time
from .hierarchy import Hierarchy


class LayoutHierarchyController:
    npu_hierarchy = None
    bench_hierarchy = None
    single_hierarchy = None

    hierarchy = {
        'NPU': npu_hierarchy,
        'Bench': bench_hierarchy,
        'Single': single_hierarchy
    }

    @staticmethod
    def change_expand_state(node_name, graph_type, graph, micro_step):
        if node_name == 'root':
            LayoutHierarchyController.hierarchy[graph_type] = Hierarchy(graph_type, graph, micro_step)
        elif LayoutHierarchyController.hierarchy[graph_type]:
            LayoutHierarchyController.hierarchy[graph_type].update_graph_data(node_name, graph)
            LayoutHierarchyController.hierarchy[graph_type].update_graph_shape()
            LayoutHierarchyController.hierarchy[graph_type].update_graph_position()
        else:
            return {}
        return LayoutHierarchyController.hierarchy[graph_type].get_hierarchy()

    @staticmethod
    def update_hierarchy_data(graph_type):
        if LayoutHierarchyController.hierarchy.get(graph_type, None):
            return LayoutHierarchyController.hierarchy[graph_type].update_hierarchy_data()
        else:
            return {}



# Copyright (c) 2025, Huawei Technologies.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# 全局常量
ADD_MATCH_KEYS = [
    'MaxAbsErr',
    'MinAbsErr',
    'MeanAbsErr',
    'NormAbsErr',
    'MaxRelativeErr',
    'MinRelativeErr',
    'MeanRelativeErr',
    'NormRelativeErr',
]
MAX_FILE_SIZE = 3 * 1024 * 1024 * 1024  # 最大文件大小限制
NPU_PREFIX = 'N___'
BENCH_PREFIX = 'B___'
FILE_NAME_REGEX = r'^[a-zA-Z0-9_\-\.]+$'  # 文件名正则表达式
# 图类型
NPU = 'NPU'
BENCH = 'Bench'
SINGLE = 'Single'
# 前端节点类型
EXPAND_MODULE = 0
UNEXPAND_NODE = 1

# 后端节点类型
MODULE = 0
API = 1
MULTI_COLLECTION = 8
API_LIST = 9


class GraphState:
    # 模块级全局变量
    _state = {
        'logdir': '',
        'current_tag': '',
        'current_run': '',
        'current_file_path': '',
        'current_file_data': {},
        'current_hierarchy': {},
        'config_data': {
            "run": "",
            "npuUnMatchNodes": [],
            "benchUnMatchNodes": [],
            "npuMatchNodes": {},
            "benchMatchNodes": {},
            "manualMatchNodes": {},
            
        },
        'first_run_tags': {},
        'runs': {},
    }

    @staticmethod
    def init_defaults():
        """
        初始化全局变量的默认值
        """
        global _state
        GraphState._state = {
            'logdir': '',
            'current_tag': '',
            'current_run': '',
            'current_file_path': '',
            'current_file_data': {},
            'current_hierarchy': {},
            'config_data': {
                "run": "",
                "npuUnMatchNodes": [],
                "benchUnMatchNodes": [],
                "npuMatchNodes": {},
                "benchMatchNodes": {},
                "manualMatchNodes": {},
                
            },
            'first_run_tags': {},
            'runs': {},
        }

    @staticmethod
    def set_global_value(key, value):
        """
        设置全局变量的值
        """
        GraphState._state[key] = value

    @staticmethod
    def get_global_value(key, default=None):
        """
        获取全局变量的值
        """
        return GraphState._state.get(key, default)

    @staticmethod
    def reset_global_state():
        """
        重置所有全局变量为默认值
        """
        GraphState.init_defaults()



# Copyright (c) 2025, Huawei Technologies.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from ..utils.graph_utils import GraphUtils
from ..utils.global_state import ADD_MATCH_KEYS, MODULE
from ..utils.global_state import GraphState


class MatchNodesController:

    @staticmethod
    def is_same_node_type(graph_data, npu_node_name, bench_node_name):
        npu_node_type = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name, {}).get('node_type')
        bench_node_type = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name, {}).get('node_type')
        if npu_node_type is None or bench_node_type is None or npu_node_type != bench_node_type:
            return False
        return True

    @staticmethod
    def process_task_add(graph_data, npu_node_name, bench_node_name, task):
        result = {}
        if task == 'md5':
            result = MatchNodesController.process_md5_task_add(graph_data, npu_node_name, bench_node_name)
        elif task == 'summary':
            result = MatchNodesController.process_summary_task_add(graph_data, npu_node_name, bench_node_name)
        else:
            result = {
                'success': False,
                'error': 'task类型错误'
            }
        return result

    @staticmethod
    def process_task_delete(graph_data, npu_node_name, bench_node_name, task):
        result = {}
        if task == 'md5':
            result = MatchNodesController.process_md5_task_delete(graph_data, npu_node_name, bench_node_name)
        elif task == 'summary':
            result = MatchNodesController.process_summary_task_delete(graph_data, npu_node_name, bench_node_name)
        else:
            result = {
                'success': False,
                'error': 'task类型错误'
            }
        return result

    @staticmethod
    def process_task_add_child_layer_by_config(graph_data, match_node_links, task):
        # 根据配置文件中的匹配关系，批量调用 process_task_add
        result = {}
        match_reslut = []
        for npu_node_name, bench_node_name in match_node_links.items():
            res = MatchNodesController.process_task_add(graph_data, npu_node_name, bench_node_name, task)
            match_reslut.append(res.get('success'))

        config_data = GraphState.get_global_value("config_data")
        
        result['success'] = True
        result['data'] = {
            'matchReslut': match_reslut,
            'npuMatchNodes': config_data.get('npuMatchNodes', {}),
            'benchMatchNodes': config_data.get('benchMatchNodes', {}),
            'npuUnMatchNodes': config_data.get('npuUnMatchNodes', []),
            'benchUnMatchNodes': config_data.get('benchUnMatchNodes', [])
        }
        return result

    @staticmethod
    def process_task_add_child_layer(graph_data, npu_node_name, bench_node_name, task):
        if not all([graph_data, npu_node_name, bench_node_name, task]):
            return {'success': False, 'error': '参数错误'}
     
        if not MatchNodesController.is_same_node_type(graph_data, npu_node_name, bench_node_name):
            return {
                'success': False,
                'error': '节点类型不一致,无法添加匹配关系'
            }
        npu_nodes = graph_data.get('NPU', {}).get('node', {})
        bench_nodes = graph_data.get('Bench', {}).get('node', {})
        # 1. 选中的目标节点和标杆侧节点添加匹配关系
        result = MatchNodesController.process_task_add(graph_data, npu_node_name, bench_node_name, task)

        # 2. 目标节点的子节点和标杆侧的子节点添加匹配关系

        def process_child_layer(npu_subnodes, bench_subnodes):
            npu_match_names = {}
            bench_match_names = {}
            # 2.1 提取目标侧节点的未匹配子节点名称和标杆侧节点的未匹配子节点名称
            for npu_subnode_name in npu_subnodes:
                node_info = npu_nodes.get(npu_subnode_name, {})
                if node_info.get('matched_node_link'):
                    continue
                node_type = node_info.get('node_type') if node_info else None
                # 2.2 处理节点名称，提取匹配信息
                match_name = extract_module_name(npu_subnode_name) if node_type == MODULE else extract_api_name(
                    npu_subnode_name)
                matched_name = f'{match_name}.{node_type}'
                npu_match_names.setdefault(matched_name, []).append(npu_subnode_name)

            for bench_subnode_name in bench_subnodes:
                node_info = bench_nodes.get(bench_subnode_name, {})
                if node_info.get('matched_node_link'):
                    continue
                node_type = node_info.get('node_type') if node_info else None
                match_name = extract_module_name(bench_subnode_name) if node_type == MODULE else extract_api_name(
                    bench_subnode_name)
                matched_name = f'{match_name}.{node_type}'
                bench_match_names.setdefault(matched_name, []).append(bench_subnode_name)

            common_keys = npu_match_names.keys() & bench_match_names.keys()
            # 2.3 取名称交集，添加匹配关系
            for key in common_keys:
                npu_subnode_list = npu_match_names.get(key, [])
                bench_subnode_list = bench_match_names.get(key, [])
                
                # 多个节点可能有一个module name
                for npu_subnode_name, bench_subnode_name in zip(npu_subnode_list, bench_subnode_list):
                    result = MatchNodesController.process_task_add(graph_data, npu_subnode_name, bench_subnode_name,
                                                                   task)
                    npu_subnodes = npu_nodes.get(npu_subnode_name, {}).get('subnodes', [])
                    bench_subnodes = bench_nodes.get(bench_subnode_name, {}).get('subnodes', [])
                    # 2.4 如果有子节点，递归调用2.1-2.4
                    if result.get('success') and npu_subnodes and bench_subnodes:
                        process_child_layer(npu_subnodes, bench_subnodes)

        def extract_module_name(subnode_name):
            splited_subnode_name = subnode_name.split('.')
            if len(splited_subnode_name) < 4:
                return ""
            module_name = splited_subnode_name[-4] if not splited_subnode_name[-4].isdigit() else splited_subnode_name[
                -5]
            return module_name

        def extract_api_name(subnode_name):
            splited_subnode_name = subnode_name.split('.')
            if len(splited_subnode_name) < 2:
                return ""
            api_name = splited_subnode_name[-2] if not splited_subnode_name[-2].isdigit() else splited_subnode_name[-3]
            return api_name

        npu_subnodes = npu_nodes.get(npu_node_name, {}).get('subnodes', [])
        bench_subnodes = bench_nodes.get(bench_node_name, {}).get('subnodes', [])
        if result.get('success') and npu_subnodes and bench_subnodes:
            process_child_layer(npu_subnodes, bench_subnodes)
        if result.get('success'):
            config_data = GraphState.get_global_value("config_data")
            result['data'] = {
                'npuMatchNodes': config_data.get('npuMatchNodes', {}),
                'benchMatchNodes': config_data.get('benchMatchNodes', {}),
                'npuUnMatchNodes': config_data.get('npuUnMatchNodes', []),
                'benchUnMatchNodes': config_data.get('benchUnMatchNodes', [])
            }
        return result

    @staticmethod
    def process_task_delete_child_layer(graph_data, npu_node_name, bench_node_name, task):
        if not all([graph_data, npu_node_name, bench_node_name, task]):
            return {'success': False, 'error': '参数错误'}

        npu_nodes = graph_data.get('NPU', {}).get('node', {})
        bench_nodes = graph_data.get('Bench', {}).get('node', {})
        # 1. 选中的目标节点和标杆侧节点添加匹配关系
        result = MatchNodesController.process_task_delete(graph_data, npu_node_name, bench_node_name, task)

        # 2. 目标节点的子节点和标杆侧的子节点添加匹配关系
        def process_child_layer(npu_child_nodes):
            for npu_subnode_name in npu_child_nodes:
                npu_subnode_info = npu_nodes.get(npu_subnode_name, {})
                matched_node_link = npu_subnode_info.get('matched_node_link', [])
                if not matched_node_link:
                    continue
                bench_subnode_name = matched_node_link[-1]
                result = MatchNodesController.process_task_delete(graph_data, npu_subnode_name, bench_subnode_name,
                                                                  task)
                npu_subnodes = npu_nodes.get(npu_subnode_name, {}).get('subnodes', [])
                bench_subnodes = bench_nodes.get(bench_subnode_name, {}).get('subnodes', [])
                # 2.4 如果有子节点，递归调用2.1-2.4
                if result.get('success') and npu_subnodes and bench_subnodes:
                    process_child_layer(npu_subnodes)

        npu_subnodes = npu_nodes.get(npu_node_name, {}).get('subnodes', [])
        bench_subnodes = bench_nodes.get(bench_node_name, {}).get('subnodes', [])
        if result.get('success') and npu_subnodes and bench_subnodes:
            process_child_layer(npu_subnodes)
        if result.get('success'):
            config_data = GraphState.get_global_value("config_data")
            
            result['data'] = {
                'npuMatchNodes': config_data.get('npuMatchNodes', {}),
                'benchMatchNodes': config_data.get('benchMatchNodes', {}),
                'npuUnMatchNodes': config_data.get('npuUnMatchNodes', []),
                'benchUnMatchNodes': config_data.get('benchUnMatchNodes', [])
            }
        return result

    @staticmethod
    def process_md5_task_add(graph_data, npu_node_name, bench_node_name):
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name, {})
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name, {})
        # 去除节点名称前缀
        npu_input_data = GraphUtils.remove_prefix(npu_node_data.get('input_data', {}), npu_node_name + '.')
        bench_input_data = GraphUtils.remove_prefix(bench_node_data.get('input_data', {}), bench_node_name + '.')
        npu_output_data = GraphUtils.remove_prefix(npu_node_data.get('output_data', {}), npu_node_name + '.')
        bench_output_data = GraphUtils.remove_prefix(bench_node_data.get('output_data', {}), bench_node_name + '.')
        # 计算精度误差
        precision_input_error = MatchNodesController.calculate_md5_diff(npu_input_data, bench_input_data)
        precision_output_error = MatchNodesController.calculate_md5_diff(npu_output_data, bench_output_data)
        precision_error = precision_input_error and precision_output_error
        # 在原始数据上，添加匹配节点，和匹配节点信息

        npu_graph_data = graph_data.get('NPU', {})
        bench_graph_data = graph_data.get('Bench', {})
        npu_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(bench_graph_data, bench_node_name)
        bench_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(npu_graph_data, npu_node_name)
        npu_node_data.setdefault('data', {})['precision_index'] = precision_error
        
        MatchNodesController.add_config_match_nodes(npu_node_name, bench_node_name)
        return {'success': True}

    @staticmethod
    def process_summary_task_add(graph_data, npu_node_name, bench_node_name):
        # 节点信息提取
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name)
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name)
        # 计算统计误差
        intput_statistical_diff = MatchNodesController.calculate_statistical_diff(
            npu_node_data.get('input_data'), bench_node_data.get('input_data'), npu_node_name, bench_node_name
        )
        output_statistical_diff = MatchNodesController.calculate_statistical_diff(
            npu_node_data.get('output_data'), bench_node_data.get('output_data'), npu_node_name, bench_node_name
        )
        # 计算精度误差
        precision_error = MatchNodesController.calculate_max_relative_error(output_statistical_diff)
        # 有一个没有匹配上，则认为匹配失败
        if not intput_statistical_diff or not output_statistical_diff:
            return {
                'success': False,
                'error': '输入或输出统计误差值为空(Input and output statistical error calculation failed)',
            }

        if precision_error == -1:
            return {
                'success': False,
                'error': '输出统计误差值为空，计算精度误差失败(Calculation of precision error failed)',
            }
        # 在原始数据上，添加匹配节点，和匹配节点信息
        npu_graph_data = graph_data.get('NPU', {})
        bench_graph_data = graph_data.get('Bench', {})
        npu_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(bench_graph_data, bench_node_name)
        bench_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(npu_graph_data, npu_node_name)
        npu_node_data.setdefault('data', {})['precision_index'] = precision_error
        MatchNodesController.update_graph_node_data(npu_node_data.get('input_data'), intput_statistical_diff)
        MatchNodesController.update_graph_node_data(npu_node_data.get('output_data'), output_statistical_diff)
        MatchNodesController.add_config_match_nodes(npu_node_name, bench_node_name)
        return {'success': True}

    @staticmethod
    def process_md5_task_delete(graph_data, npu_node_name, bench_node_name):
        config_data = GraphState.get_global_value("config_data")
        manual_match_nodes = config_data.get('manualMatchNodes', {})
        if manual_match_nodes.get(npu_node_name) != bench_node_name:
            return {
                'success': False,
                'error': "操作失败：节点未匹配，请先匹配节点",
            }
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name, {})
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name, {})
        # 在原始数据上，删除匹配节点，和匹配节点信息
        npu_node_data['matched_node_link'] = []
        bench_node_data['matched_node_link'] = []
        # 后端维护一个匹配节点列表，前端展示
        del npu_node_data['data']['precision_index']
        MatchNodesController.delete_config_match_nodes(npu_node_name, bench_node_name)
        return {
            'success': True,
            'data': {},
        }

    @staticmethod
    def process_summary_task_delete(graph_data, npu_node_name, bench_node_name):
        config_data = GraphState.get_global_value("config_data")
        manual_match_nodes = config_data.get('manualMatchNodes', {})
        if manual_match_nodes.get(npu_node_name) != bench_node_name:
            return {
                'success': False,
                'error': "操作失败：节点未匹配，请先匹配节点",
            }
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name, {})
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name, {})
        # 在原始数据上，删除匹配节点，和匹配节点信息
        npu_node_data['matched_node_link'] = []
        bench_node_data['matched_node_link'] = []
        MatchNodesController.delete_matched_node_data(npu_node_data.get('input_data'))
        MatchNodesController.delete_matched_node_data(npu_node_data.get('output_data'))
        # 防止 KeyError 或 TypeError
        npu_node_data.get('data', {}).pop('precision_index', None)
        MatchNodesController.delete_config_match_nodes(npu_node_name, bench_node_name)
        return {
            'success': True,
            'data': {},
        }

    @staticmethod
    def add_config_match_nodes(npu_node_name, bench_node_name):
        config_data = GraphState.get_global_value("config_data")
        # 匹配列表和未匹配列表
        manual_match_nodes = config_data.setdefault('manualMatchNodes', {})
        npu_match_nodes_list = config_data.setdefault('npuMatchNodes', {})
        bench_match_nodes_list = config_data.setdefault('benchMatchNodes', {})
        npu_unmatehed_name_list = config_data.setdefault('npuUnMatchNodes', [])
        bench_unmatehed_name_list = config_data.setdefault('benchUnMatchNodes', [])
        # 更新匹配列表和未匹配列表
        if str(npu_node_name) in npu_unmatehed_name_list:
            npu_unmatehed_name_list.remove(str(npu_node_name))
        if str(bench_node_name) in bench_unmatehed_name_list:
            bench_unmatehed_name_list.remove(str(bench_node_name))
        manual_match_nodes[str(npu_node_name)] = str(bench_node_name)
        npu_match_nodes_list[str(npu_node_name)] = str(bench_node_name)
        bench_match_nodes_list[str(bench_node_name)] = str(npu_node_name)
        GraphState.set_global_value("config_data", config_data)

    @staticmethod
    def delete_config_match_nodes(npu_node_name, bench_node_name):
        config_data = GraphState.get_global_value("config_data")
        # 匹配列表和未匹配列表
        manual_match_nodes = config_data.setdefault('manualMatchNodes', {})
        npu_match_nodes_list = config_data.setdefault('npuMatchNodes', {})
        bench_match_nodes_list = config_data.setdefault('benchMatchNodes', {})
        npu_unmatehed_name_list = config_data.setdefault('npuUnMatchNodes', [])
        bench_unmatehed_name_list = config_data.setdefault('benchUnMatchNodes', [])
        # 更新匹配列表和未匹配列表
        if str(npu_node_name) in manual_match_nodes:
            del manual_match_nodes[str(npu_node_name)]
        if str(npu_node_name) in npu_match_nodes_list:
            del npu_match_nodes_list[str(npu_node_name)]
        if str(bench_node_name) in bench_match_nodes_list:
            del bench_match_nodes_list[str(bench_node_name)]
        npu_unmatehed_name_list.append(str(npu_node_name))
        bench_unmatehed_name_list.append(str(bench_node_name))
        GraphState.set_global_value("config_data", config_data)

    @staticmethod
    def calculate_statistical_diff(npu_data, bench_data, npu_node_name, bench_node_name):
        result = {}
        # 去除节点名称前缀并转化为列表形式
        npu_data_simple = list(GraphUtils.remove_prefix(npu_data, npu_node_name + '.').values())
        bench_data_simple = list(GraphUtils.remove_prefix(bench_data, bench_node_name + '.').values())
        npu_data_keys = list(GraphUtils.remove_prefix(npu_data, npu_node_name + '.').keys())
        # 使用 zip 只对比最短的列表长度
        for npu_values, bench_values in zip(npu_data_simple, bench_data_simple):
            npu_max = GraphUtils.convert_to_float(npu_values.get('Max', float('nan')))
            bench_max = GraphUtils.convert_to_float(bench_values.get('Max', float('nan')))
            npu_min = GraphUtils.convert_to_float(npu_values.get('Min', float('nan')))
            bench_min = GraphUtils.convert_to_float(bench_values.get('Min', float('nan')))
            npu_norm = GraphUtils.convert_to_float(npu_values.get('Norm', float('nan')))
            bench_norm = GraphUtils.convert_to_float(bench_values.get('Norm', float('nan')))
            npu_mean = GraphUtils.convert_to_float(npu_values.get('Mean', float('nan')))
            bench_mean = GraphUtils.convert_to_float(bench_values.get('Mean', float('nan')))

            # Calculate absolute differences
            max_diff_abs = abs(npu_max - bench_max)
            min_diff_abs = abs(npu_min - bench_min)
            norm_diff_abs = abs(npu_norm - bench_norm)
            mean_diff_abs = abs(npu_mean - bench_mean)

            # Calculate relative differences (avoid division by zero)
            max_diff_rel = abs(max_diff_abs / (bench_max if bench_max != 0 else 1))
            min_diff_rel = abs(min_diff_abs / (bench_min if bench_min != 0 else 1))
            norm_diff_rel = abs(norm_diff_abs / (bench_norm if bench_norm != 0 else 1))
            mean_diff_rel = abs(mean_diff_abs / (bench_mean if bench_mean != 0 else 1))

            # 将结果记录到字典中
            result[npu_node_name + '.' + npu_data_keys[len(result)]] = dict(
                zip(
                    ADD_MATCH_KEYS,
                    [
                        max_diff_abs,
                        min_diff_abs,
                        mean_diff_abs,
                        norm_diff_abs,
                        max_diff_rel,
                        min_diff_rel,
                        mean_diff_rel,
                        norm_diff_rel,
                    ],
                )
            )

        return result

    # 计算最大相对误差
    @staticmethod
    def calculate_max_relative_error(result):
        max_rel_error = -1
        for _, diff_values in result.items():
            max_diff_rel = diff_values.get('MaxRelativeErr', float('nan'))
            min_diff_rel = diff_values.get('MinRelativeErr', float('nan'))
            norm_diff_rel = diff_values.get('NormRelativeErr', float('nan'))
            mean_diff_rel = diff_values.get('MeanRelativeErr', float('nan'))

            max_rel_error_for_key = max(max_diff_rel, min_diff_rel, norm_diff_rel, mean_diff_rel)
            max_rel_error = max(max_rel_error, max_rel_error_for_key)

        return min(max_rel_error, 1)

    @staticmethod
    def calculate_md5_diff(npu_data, bench_data):
        if npu_data == {} or bench_data == {}:
            return 0
        # 对比每个NPU和Bench所有数据md值，如果有一个不一样则返回0,否则返回1
        for npu_key, bench_key in zip(npu_data, bench_data):
            npu_md5 = npu_data[npu_key].get('md5', '')
            bench_md5 = bench_data[bench_key].get('md5', '')
            if npu_md5 != bench_md5:
                return 0
        return 1

    @staticmethod
    def update_graph_node_data(graph_npu_node_data, statistical_diff):
        if not statistical_diff or not graph_npu_node_data:
            return
        for key, diff_values in statistical_diff.items():
            # 格式化相对误差字段
            for field in ['MaxRelativeErr', 'MinRelativeErr', 'NormRelativeErr', 'MeanRelativeErr']:
                diff_values[field] = GraphUtils.format_relative_err(diff_values.get(field, float('nan')))

            # 转换 absErr 为 NaN 字符串
            for field in ['MaxAbsErr', 'MinAbsErr', 'MeanAbsErr', 'NormAbsErr']:
                diff_values[field] = GraphUtils.nan_to_str(diff_values.get(field, float('nan')))
            if key in graph_npu_node_data:
                graph_npu_node_data[key].update(diff_values)
            else:
                graph_npu_node_data[key] = diff_values

    @staticmethod
    def delete_matched_node_data(graph_npu_node_data):
        keys_to_remove = ADD_MATCH_KEYS
        # 遍历graph_npu_node_data中的每个主键和对应的子字典
        for key, fild_obj in graph_npu_node_data.items():
            if not fild_obj or not isinstance(fild_obj, dict):
                continue
            # 使用字典解析创建新的子字典，排除不需要的键
            graph_npu_node_data[key] = {
                sub_key: value
                for sub_key, value in fild_obj.items()
                if sub_key not in keys_to_remove
            }



# Copyright (c) 2025, Huawei Technologies.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import time
import json
from tensorboard.util import tb_logging
from ..utils.graph_utils import GraphUtils
from ..utils.global_state import GraphState
from ..controllers.match_nodes_controller import MatchNodesController
from ..controllers.layout_hierarchy_controller import LayoutHierarchyController
from ..utils.global_state import NPU_PREFIX, BENCH_PREFIX, NPU, BENCH, SINGLE

logger = tb_logging.get_logger()


class GraphService:

    @staticmethod
    def load_meta_dir():
        """Scan logdir for directories containing .vis files, modified to return a tuple of (run, tag)."""
        logdir = GraphState.get_global_value('logdir')
        runs = GraphState.get_global_value('runs', {})
        first_run_tags = GraphState.get_global_value('first_run_tags', {})
        meta_dir = {}
        for root, _, files in GraphUtils.walk_with_max_depth(logdir, 2):
            for file in files:
                if file.endswith('.vis'):  # check for .vis extension
                    run_abs = os.path.abspath(root)
                    run = os.path.basename(run_abs)  # 不允许同名目录，否则有问题
                    tag = os.path.splitext(file)[0]  # Use the filename without extension as tag
                    _, error = GraphUtils.safe_load_data(run_abs, f"{tag}.vis", True)
                    if error:
                        logger.error(f'Error: File run:"{run_abs},tag:{tag}" is not accessible. Error: {error}')
                        continue
                    runs[run] = run_abs
                    meta_dir.setdefault(run, []).append(tag)
        meta_dir = GraphUtils.sort_data(meta_dir)
        for run, tags in meta_dir.items():
            first_run_tags[run] = tags[0]
        GraphState.set_global_value('runs', runs)
        GraphState.set_global_value('first_run_tags', first_run_tags)
        return meta_dir

    @staticmethod
    def load_graph_data(run_name, tag):
        runs = GraphState.get_global_value('runs')
        run = runs.get(run_name)
        is_safe, error = GraphUtils.safe_load_data(run_name, f"{tag}.vis", True)
        if (not is_safe or not run):
            logger.error(f'Error: File run:"{run},tag:{tag}" is not accessible. Error: {error}')
            yield f"data: {{\"error\": 读取文件失败}}\n\n"
            return
        buffer = ""
        read_bytes = 0
        chunk_size = 1024 * 1024 * 60  # 缓冲区
        json_data = None  # 最终存储的变量
        file_path = os.path.join(run, f"{tag}.vis")
        file_path = os.path.normpath(file_path)  # 标准化路径
        file_size = os.path.getsize(file_path)
        with open(file_path, 'r', encoding='utf-8') as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                read_bytes += len(chunk)
                buffer += chunk
                current_progress = min(95, int((read_bytes / file_size) * 100))
                reading_info = {
                    'progress': current_progress,
                    'status': 'reading',
                    'size': file_size,
                    'read': read_bytes
                }
                yield f"data: {json.dumps(reading_info)}\n\n"
                time.sleep(0.01)  # 避免发送过快

        if json_data is None and buffer:  # 最终验证数据
            try:
                yield f"data: {json.dumps({'progress': current_progress, 'status': 'loading'})}\n\n"
                json_data = GraphUtils.safe_json_loads(buffer)
                yield f"data: {json.dumps({'progress': 99, 'status': 'loading'})}\n\n"
            except json.JSONDecodeError as e:
                yield f"data: {json.dumps({'progress': current_progress, 'error': str(e)})}\n\n"

        if json_data is not None:  # 验证存储
            GraphState.set_global_value('current_file_data', json_data)
            GraphState.set_global_value('current_tag', tag)
            GraphState.set_global_value('current_run', run)
            yield f"data: {json.dumps({'done': True, 'progress': 100, 'status': 'loading'})}\n\n"
        else:
            yield f"data: {json.dumps({'progress': current_progress, 'error': 'Failed to parse JSON'})}\n\n"

    @staticmethod
    def load_graph_config_info(run, tag):
        graph_data, error_message = GraphUtils.get_graph_data({'run': run, 'tag': tag})
        if error_message or not graph_data:
            return {'success': False, 'error': error_message}
        config = {}
        try:

            # 读取全局信息,tag层面
            if graph_data.get('MicroSteps', {}):
                config['microSteps'] = graph_data.get('MicroSteps')
            if graph_data.get('ToolTip', {}):
                config['tooltips'] = graph_data.get('ToolTip')
            config['overflowCheck'] = bool(graph_data.get('OverflowCheck')) if 'OverflowCheck' in graph_data else True
            config['isSingleGraph'] = False if graph_data.get(NPU) else True
            # 读取配置信息，run层面
            config_data = GraphState.get_global_value("config_data", {})
            config_data_run = config_data.get(run, {})
            if not config_data_run:  # 如果没有run的配置信息，则读取第一个文件中的Colors
                first_run_tags = GraphState.get_global_value("first_run_tags")
                first_tag = first_run_tags.get(run)
                if not first_tag:
                    return {'success': False, 'error': '获取配置信息失败,请检查目录中第一个文件'}
                first_graph_data, error_message = GraphUtils.get_graph_data({'run': run, 'tag': first_tag})
                config_data_run['colors'] = first_graph_data.get('Colors')
                config_data[run] = config_data_run
                GraphState.set_global_value('config_data', config_data)
                config['colors'] = first_graph_data.get('Colors')
            else:
                config['colors'] = config_data_run.get('colors')
            # 读取目录下配置文件列表
            config_files = GraphUtils.find_config_files(run)
            config['matchedConfigFiles'] = config_files
            config['task'] = graph_data.get('task')
            return {'success': True, 'data': config}
        except Exception as e:
            return {'success': False, 'error': '获取配置信息失败,请检查目录中第一个文件'}

    @staticmethod
    def load_graph_all_node_list(run, tag, micro_step):
        graph_data, error_message = GraphUtils.get_graph_data({'run': run, 'tag': tag})
        if error_message or not graph_data:
            return {'success': False, 'error': error_message}
        result = {}
        try:
            if not graph_data.get(NPU):
                nodes = GraphUtils.split_graph_data_by_microstep(graph_data, micro_step)
                node_name_list = list(nodes.keys())
                result['npuNodeList'] = node_name_list
                return {'success': True, 'data': result}

            else:
                # 双图
                config_data = GraphState.get_global_value("config_data")
                npu_node = GraphUtils.split_graph_data_by_microstep(graph_data.get(NPU), micro_step)
                bench_node = GraphUtils.split_graph_data_by_microstep(graph_data.get(BENCH), micro_step)
                npu_node_name_list = list(npu_node.keys())
                bench_node_name_list = list(bench_node.keys())
                npu_unmatehed_name_list = [
                    key 
                    for key, value in npu_node.items() 
                    if not value.get("matched_node_link")
                ]
                bench_unmatehed_name_list = [
                    key 
                    for key, value in bench_node.items() 
                    if not value.get("matched_node_link")
                ]
                # 保存未匹配和已匹配的节点到全局变量
                config_data['npuUnMatchNodes'] = npu_unmatehed_name_list
                config_data['benchUnMatchNodes'] = bench_unmatehed_name_list
                config_data['npuMatchNodes'] = graph_data.setdefault('npu_match_nodes', {})
                config_data['benchMatchNodes'] = graph_data.setdefault('bench_match_nodes', {})
                # 返回结果
                result['npuNodeList'] = npu_node_name_list
                result['benchNodeList'] = bench_node_name_list
                result['npuUnMatchNodes'] = npu_unmatehed_name_list
                result['benchUnMatchNodes'] = bench_unmatehed_name_list
                result['npuMatchNodes'] = graph_data.setdefault('npu_match_nodes', {})
                result['benchMatchNodes'] = graph_data.setdefault('bench_match_nodes', {})
                for npu_node_name, npu_node in graph_data.get(NPU, {}).get('node', {}).items():
                    if npu_node.get('matched_node_link', None):
                        bench_node_name = npu_node.get('matched_node_link', [None])[-1].replace(BENCH_PREFIX, '', 1)
                        result['npuMatchNodes'][npu_node_name] = bench_node_name
                        result['benchMatchNodes'][bench_node_name] = npu_node_name
                
                GraphState.set_global_value('config_data', config_data)
                return {'success': True, 'data': result}
        except Exception as e:
            logger.error('获取节点列表失败:' + str(e))
            return {'success': False, 'error': '获取节点列表失败:' + str(e)}

    @staticmethod
    def change_node_expand_state(node_info, meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message or not graph_data:
            return {'success': False, 'error': error_message}
        graph_type = node_info.get('nodeType')
        node_name = node_info.get('nodeName')
        micro_step = meta_data.get('microStep')
        try:
            # 单图
            if not graph_data.get(NPU):
                hierarchy = LayoutHierarchyController.change_expand_state(node_name, SINGLE, graph_data, micro_step)
            # NPU
            elif (graph_type == NPU):
                hierarchy = LayoutHierarchyController.change_expand_state(node_name, graph_type,
                                                                          graph_data.get(NPU, {}), micro_step)
            # 标杆
            elif graph_type == BENCH:
                hierarchy = LayoutHierarchyController.change_expand_state(node_name, graph_type,
                                                                          graph_data.get(BENCH, {}), micro_step)
            else:
                return {'success': True, 'data': {}}
            return {'success': True, 'data': hierarchy}
        except Exception as e:
            logger.error('节点展开或收起发生错误:' + str(e))
            node_type_name = ""
            if graph_data.get(NPU): 
                node_type_name = '调试侧' if graph_type == NPU else '标杆侧'
            return {'success': False, 'error': f'{node_type_name}节点展开或收起发生错误', 'data': None}

    @staticmethod
    def update_hierarchy_data(graph_type):
        if (graph_type == NPU or graph_type == BENCH):
            hierarchy = LayoutHierarchyController.update_hierarchy_data(graph_type)
            return {'success': True, 'data': hierarchy}
        else:
            return {'success': False, 'error': '节点类型错误'}

    @staticmethod
    def get_node_info(node_info, meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': error_message}
        try:
            graph_type = node_info.get('nodeType')
            node_name = node_info.get('nodeName')
            result = {}
            if not graph_data.get(NPU) or graph_type == SINGLE:
                result['npu'] = graph_data.get('node', {}).get(node_name)
            else:
                matched_node_type = BENCH if graph_type == NPU else NPU
                matched_node_preifx = NPU_PREFIX if matched_node_type == NPU else BENCH_PREFIX
                node = graph_data.get(graph_type, {}).get('node', {}).get(node_name, {})
                matched_node_link = node.get('matched_node_link', []) if isinstance(node.get('matched_node_link', []),
                                                                                    list) else []
                matched_node_name = matched_node_link[-1].replace(matched_node_preifx, '',
                                                                  1) if matched_node_link else None
                matched_node = graph_data.get(matched_node_type, {}).get('node', {}).get(
                    matched_node_name) if matched_node_name else None
                result['npu'] = node if graph_type == NPU else matched_node
                result['bench'] = node if graph_type == BENCH else matched_node
            return {'success': True, 'data': result}
        except Exception as e:
            logger.error('获取节点信息失败:' + str(e))
            return {'success': False, 'error': '获取节点信息失败:' + str(e), 'data': None}

    @staticmethod
    def add_match_nodes(npu_node_name, bench_node_name, meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': error_message}
        task = graph_data.get('task')
        try:
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                result = MatchNodesController.process_task_add_child_layer(graph_data,
                                                                npu_node_name, bench_node_name, task)
                return result
            else:
                return {'success': False, 'error': '任务类型不支持(Task type not supported) '}
        except Exception as e:
            return {'success': False, '操作失败': str(e), 'data': None}

    @staticmethod
    def add_match_nodes_by_config(config_file, meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': '读取文件失败'}
        match_node_links, error = GraphUtils.safe_load_data(meta_data.get('run'), config_file)
        if error:
            return {'success': False, 'error': '配置文件失败'}
        task = graph_data.get('task')
        try:
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                result = MatchNodesController.process_task_add_child_layer_by_config(graph_data, match_node_links, task)
                return result
            else:
                return {'success': False, 'error': '任务类型不支持(Task type not supported) '}
        except Exception as e:
            return {'success': False, 'error': '操作失败', 'data': None}

    @staticmethod
    def delete_match_nodes(npu_node_name, bench_node_name, meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': error_message}
        task = graph_data.get('task')
        try:
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                result = MatchNodesController.process_task_delete_child_layer(graph_data, npu_node_name,
                                                                              bench_node_name, task)
                return result
            else:
                return {'success': False, 'error': '任务类型不支持(Task type not supported) '}
        except Exception as e:
            return {'success': False, '操作失败': str(e), 'data': None}

    @staticmethod
    def update_colors(run, colors):
        """Set new colors in jsondata."""
        try:
            config_data = GraphState.get_global_value("config_data", {})
            first_run_tags = GraphState.get_global_value("first_run_tags")
            config_data_run = config_data.get(run, {})
            first_run_tag = first_run_tags.get(run)
            first_file_data, error = GraphUtils.safe_load_data(run, f"{first_run_tag}.vis")
            if error:
                return {'success': False, 'error': '获取配置信息失败,请检查目录中第一个文件'}
            first_file_data['Colors'] = colors
            config_data_run['colors'] = colors
            config_data[run] = config_data_run
            GraphState.set_global_value("config_data", config_data)
            GraphUtils.safe_save_data(first_file_data, run, first_run_tag)
            return {'success': True, 'error': None, 'data': {}}
        except Exception as e:
            return {'success': False, 'error': str(e), 'data': None}

    @staticmethod
    def save_data(meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': error_message}

        run = meta_data.get('run')
        tag = meta_data.get('tag')
        try:
            _, error = GraphUtils.safe_save_data(graph_data, run, f"{tag}.vis")
            if error:
                return {'success': False, 'error': error}
        except (ValueError, IOError, PermissionError) as e:
            return {'success': False, 'error': f"Error: {e}"}
        return {'success': True}

    @staticmethod
    def save_matched_relations(meta_data):
        config_data = GraphState.get_global_value("config_data")
        # 匹配列表和未匹配列表
        npu_match_nodes_list = config_data.get('manualMatchNodes', {})
        run = meta_data.get('run')
        tag = meta_data.get('tag')
        try:
            _, error = GraphUtils.safe_save_data(npu_match_nodes_list, run, f"{tag}.vis.config")
            if error:
                return {'success': False, 'error': error}
        except (ValueError, IOError, PermissionError) as e:
            return {'success': False, 'error': f"Error: {e}"}
        return {'success': True, 'data': f"{tag}.vis.config"}



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (c) 2025, Huawei Technologies.
# Adapt to the model hierarchical visualization data collected by the msprobe tool
# ==============================================================================
"""The TensorBoard Graphs plugin."""

import os
from werkzeug import wrappers
from tensorboard.backend import http_util
from tensorboard.plugins import base_plugin
from tensorboard.util import tb_logging

from . import constants
from .app.views.graph_views import GraphView
from .app.utils.graph_utils import GraphUtils
from .app.utils.global_state import GraphState

logger = tb_logging.get_logger()

PLUGIN_NAME = 'graph_ascend'
PLUGIN_NAME_RUN_METADATA_WITH_GRAPH = 'graph_ascend_run_metadata_graph'


class GraphsPlugin(base_plugin.TBPlugin):
    """Graphs Plugin for TensorBoard."""

    plugin_name = PLUGIN_NAME
    headers = [('X-Content-Type-Options', 'nosniff')]

    def __init__(self, context):
        """Instantiates GraphsPlugin via TensorBoard core.

        Args:
          context: A base_plugin.TBContext instance.
        """
        super().__init__(context)
        self._data_provider = context.data_provider
        self.logdir = os.path.abspath(os.path.expanduser(context.logdir.rstrip('/')))
        # 将logdir赋值给global_state中的logdir属性,方便其他模块使用
        GraphState.set_global_value('logdir', os.path.abspath(os.path.expanduser(context.logdir.rstrip('/'))))
        self._current_file_path = None  # Store the path of the currently loaded file
        self._current_file_data = None  # Store the data of the currently loaded file
        self._current_tag = None  # Store the tag of the currently loaded file
        self.batch_id = 0  # 将 batch_id 声明为实例变量
        self.step_id = 0  # 可以同样声明 step_id
        self.dfs_node_ids = []  # batch和step没变的话就将所有的nodename存起来，方便快速读取
        self.check_batch_id = -1  # 来配合node_ids监察用的，他不变node_ids就不用重新读取了
        self.check_step_id = 0  # 同上
        self.check_tag = None

    def get_plugin_apps(self):
        return {
            '/index.js': GraphView.static_file_route,
            '/index.html': GraphView.static_file_route,
            "/load_meta_dir": GraphView.load_meta_dir,
            "/screen": self.get_all_screen_nodes,
            '/loadGraphData': GraphView.load_graph_data,
            '/loadGraphConfigInfo': GraphView.load_graph_config_info,
            '/loadGraphAllNodeList': GraphView.load_graph_all_node_list,
            '/changeNodeExpandState': GraphView.change_node_expand_state,
            '/updateHierarchyData': GraphView.update_hierarchy_data,
            '/getNodeInfo': GraphView.get_node_info,
            '/addMatchNodes': GraphView.add_match_nodes,
            '/addMatchNodesByConfig': GraphView.add_match_nodes_by_config,
            '/deleteMatchNodes': GraphView.delete_match_nodes,
            '/saveData': GraphView.save_data,
            '/updateColors': GraphView.update_colors,
            '/saveMatchedRelations': GraphView.save_matched_relations,
        }

    def is_active(self):
        """The graphs plugin is active if any run has a graph."""
        for content in os.listdir(self.logdir):
            content_path = os.path.join(self.logdir, content)
            if os.path.isfile(content_path) and content.endswith('.vis'):
                return True
            if os.path.isdir(content_path):
                for file in os.listdir(content_path):
                    if os.path.isfile(os.path.join(content_path, file)) and file.endswith('.vis'):
                        return True
        return False

    def data_plugin_names(self):
        return (
            PLUGIN_NAME,
            PLUGIN_NAME_RUN_METADATA_WITH_GRAPH,
        )

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(
            es_module_path='/index.js',
            disable_reload=True,
        )

    # 拿所有precisonNodes的，与controls的精度筛选联动
    @wrappers.Request.application
    def get_all_screen_nodes(self, request):
        grouped_screen_set, inaccuracy_node_ids = [], []
        precision_none = 0
        screen = ''
        # 尝试获取 screen_set 和 screen 的值
        for key, value in constants.SCREEN_MAP.items():
            if key in request.args:
                screen_set = request.args.get(key)
                screen = value
                break  # 找到一个匹配的 key 后跳出循环

        if screen == 'precision_index':
            precision_set_str = screen_set.split(',')
            if constants.UNMATCHED_NODE_NAME in precision_set_str:
                precision_set_str = [p for p in precision_set_str if p != constants.UNMATCHED_NODE_NAME]
                precision_none = 1
            grouped_screen_set = [
                list(map(float, precision_set_str[i: i + 2]))
                for i in range(0, len(precision_set_str), 2)
            ]
        else:
            grouped_screen_set = screen_set
        tag = request.args.get("tag")
        json_data = self.check_jsondata(request)

        def has_conditions_changed(tag, batch):
            return (
                    self.check_batch_id != batch
                    or self.check_step_id != self.step_id
                    or self.check_tag != tag
                    or self.check_tag is None
            )

        if has_conditions_changed(tag, self.batch_id):
            self.dfs_node_ids = self.dfs_collect_nodes(json_data, request)
            self.check_batch_id = self.batch_id
            self.check_step_id = self.step_id
            self.check_tag = tag
        node_ids = self.dfs_node_ids
        for node in node_ids:
            node_data = self.json_get(json_data, 'NPU', 'node', node, 'data') or self.json_get(
                json_data, 'node', node, 'data'
            )
            matched = self.json_get(json_data, 'NPU', 'node', node, 'matched_node_link') or self.json_get(
                json_data, 'node', node, 'matched_node_link'
            )
            inaccuracy = node_data.get(screen) if node_data is not None else None
            # 如果 inaccuracy 为 None，直接检查是否符合条件
            if inaccuracy is None and precision_none == 0:
                continue  # 跳过后续的处理，进入下一个 node
            if inaccuracy is None and precision_none == 1:
                if (node_data is None or node_data.get('overflow_level', False)) and not matched:
                    inaccuracy_node_ids.append(node)
                continue  # 跳过后续的处理，进入下一个 node

            # 对于 inaccuracy 是数字类型，检查是否在某个子范围内，精度误差
            if isinstance(inaccuracy, (int, float)):
                for group in grouped_screen_set:
                    if len(group) > 1 and all(g is not None for g in group) and group[0] <= inaccuracy <= group[1]:
                        inaccuracy_node_ids.append(node)
                        break  # 找到符合条件的，跳出当前循环
            # 对于非数字的 inaccuracy，检查是否在 grouped_screen_set 中，溢出检测
            elif inaccuracy in grouped_screen_set:
                inaccuracy_node_ids.append(node)
            else:
                logger.error(f'The inaccuracy in {node} is not a valid value')

        return http_util.Respond(request, inaccuracy_node_ids, "application/json")

    def dfs_collect_nodes(self, json_data, request):
        root_subnodes_set = []
        all_node_names = []
        try:
            request_micro_step_id = request.args.get("microStep")
        except ValueError:
            logger.error('The param "batch" or "step" does not exist or not a valid value')
        root_name = self.json_get(json_data, 'NPU', 'root') or \
                    self.json_get(json_data, 'root')
        root_subnodes = self.json_get(json_data, 'NPU', 'node', root_name, 'subnodes') \
            if 'NPU' in json_data else \
            self.json_get(json_data, 'node', root_name, 'subnodes')
        if root_subnodes:
            for node in root_subnodes:
                json_path = ['NPU', 'node', node, 'micro_step_id'] if 'NPU' in json_data \
                    else ['node', node, 'micro_step_id']
                micro_step_id = self.json_get(json_data, *json_path)
                if request_micro_step_id == '-1' or str(micro_step_id) == request_micro_step_id:
                    root_subnodes_set.append(node)

        def get_leaf_nodes(subnodes_set):
            npu_data = self.json_get(json_data, 'NPU')
            for node in subnodes_set:
                node_data = (
                    self.json_get(npu_data, 'node', node) if npu_data else self.json_get(json_data, 'node', node)
                )
                if node_data:
                    if node_data.get('subnodes'):
                        get_leaf_nodes(node_data.get('subnodes'))
                    else:
                        all_node_names.append(node)

        get_leaf_nodes(root_subnodes_set)

        return all_node_names

    # 检查到底是读一般还是用之前存的
    def check_jsondata(self, request):
        meta_data = {
            "tag": request.args.get("tag"),
            "run": request.args.get('run')
        }
        graph_data, _ = GraphUtils.get_graph_data(meta_data)
        return graph_data

    def json_get(self, data, *args):
        result = data
        for key in args:
            if result is None:
                return None
            result = result.get(key)
        return result



# -------------------------------------------------------------------------
# Copyright (c) 2025, Huawei Technologies.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0  (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------------------------#
import setuptools

VERSION = '8.1.0'
INSTALL_REQUIRED = ["tensorboard >= 2.11.2"]

setuptools.setup(
    name="tb-graph-ascend",
    version=VERSION,
    description="Model Hierarchical Visualization TensorBoard Plugin",
    long_description="Model Hierarchical Visualization TensorBoard Plugin : \
        https://gitee.com/ascend/mstt/tree/master/plugins/tensorboard-plugins/tb_graph_ascend",
    url="https://gitee.com/ascend/mstt/tree/master/plugins/tensorboard-plugins/tb_graph_ascend",
    author="Ascend Team",
    author_email="pmail_mindstudio@huawei.com",
    packages=setuptools.find_packages(),
    package_data={
        "server": ["static/**"],
    },
    entry_points={
        "tensorboard_plugins": [
            "graph_ascend = server.plugin:GraphsPlugin",
        ],
    },
    python_requires=">=3.7",
    install_requires=INSTALL_REQUIRED,
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    license='BSD-3',
    keywords='tensorboard graph ascend plugin',
)



# Copyright (c) 2025, Huawei Technologies.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import math
import os
import json
import re
import stat
from functools import cmp_to_key
from pathlib import Path
from tensorboard.util import tb_logging
from .global_state import GraphState, FILE_NAME_REGEX, MAX_FILE_SIZE

logger = tb_logging.get_logger()


class GraphUtils:

    @staticmethod
    def get_graph_data(meta_data):
        if not meta_data:
            return None, 'Error: no query parameters provided'
        try:
            run_name = meta_data.get('run')
            runs = GraphState.get_global_value('runs', {})
            run = runs.get(run_name) or run_name
            tag = meta_data.get('tag')
            current_tag = GraphState.get_global_value('current_tag')
            current_run = GraphState.get_global_value('current_run')
            if current_tag == tag and current_run == run:
                return GraphState.get_global_value('current_file_data'), None  # 直接返回获取结果
            else:
                json_data, error_message = GraphUtils.safe_load_data(run, f"{tag}.vis", False)
                if error_message:
                    return None, error_message
                GraphState.set_global_value('current_file_data', json_data)
                GraphState.set_global_value('current_tag', tag)
                GraphState.set_global_value('current_run', run)
                return json_data, error_message
        except Exception as e:
            logger.error(f'Error: fail to get graph data by {meta_data}, error: {e}')
            return None, 'Error: fail to get graph data'

    @staticmethod
    def get_parent_node_list(graph_data, node_name):
        """获取父节点列表"""
        # 如果 graph_data 为空或 node_name 不存在，直接返回空列表
        if not graph_data or not node_name:
            return []

        node_list = []  # 存储结果的列表
        current_node = node_name  # 当前节点名称
        nodes = graph_data['node']  # 提取节点数据，避免重复访问

        while current_node:
            node_list.append(current_node)  # 将当前节点加入列表
            # 获取当前节点的 upnode
            current_node = nodes.get(current_node, {}).get('upnode')

            # 检测循环引用（防止死循环）
            if current_node in node_list:
                raise ValueError(f"检测到循环引用：节点 {current_node} 已存在于路径中")

        return list(reversed(node_list))  # 返回结果列表

    @staticmethod
    def split_graph_data_by_microstep(graph_data, micro_step):
        graph_nodes = graph_data.get('node', {})
        if str(micro_step) == str(-1):
            return graph_nodes
        splited_graph_data = {}

        def traverse_npu(graph_nodes, subnodes):
            for node_name in subnodes:
                node_data = graph_nodes.get(node_name, {})
                micro_step_id = node_data.get('micro_step_id')
                if str(micro_step_id) == str(micro_step) or micro_step_id is None:
                    splited_graph_data[node_name] = (node_data)
                    traverse_npu(graph_nodes, node_data.get('subnodes', []))

        root = graph_data.get('root')
        root_subnodes = graph_nodes.get(root, {}).get('subnodes', [])
        traverse_npu(graph_nodes, root_subnodes)
        return splited_graph_data

    @staticmethod
    def walk_with_max_depth(logdir, max_depth):
        for root, dirs, files in os.walk(logdir):
            # 计算当前 root 相对于 top 的深度
            depth = root[len(logdir):].count(os.sep) + 1
            if depth >= max_depth:
                del dirs[:]
            yield root, dirs, files

    @staticmethod   
    def safe_json_loads(json_str, default_value=None):
        """
        安全地解析 JSON 字符串，带长度限制和异常处理。
        :param json_str: 要解析的 JSON 字符串
        :param default_value: 如果解析失败返回的默认值
        :return: 解析后的 Python 对象 或 default_value
        """
        # 类型检查
        if not isinstance(json_str, str):
            logger.error("Input is not a string.")
            return default_value

        # 长度限制
        if len(json_str) > MAX_FILE_SIZE:
            logger.error(f"Input length exceeds {MAX_FILE_SIZE} characters.")
            return default_value

        try:
            result = json.loads(json_str)
            GraphUtils.remove_prototype_pollution(result)
            return result
        except json.JSONDecodeError as e:
            logger.error(f"JSON decode error: {e}")
            return default_value
        except Exception as e:
            logger.error(f"Unexpected error: {e}")
            return default_value

    @staticmethod   
    def remove_prototype_pollution(obj, current_depth=1, max_depth=200):
        """
        递归删除对象中的原型污染字段，如 '__proto__', 'constructor', 'prototype'。
        
        :param obj: 要清理的对象
        :param current_depth: 当前递归深度，默认从 1 开始
        :param max_depth: 最大允许递归深度
        """
        if current_depth > max_depth:
            logger.warning(f"Reached maximum recursion depth of {max_depth}. Stopping further recursion.")
            return
        
        if isinstance(obj, dict):
            for key in list(obj.keys()):
                if key in ('__proto__', 'constructor', 'prototype'):
                    del obj[key]
                else:
                    GraphUtils.remove_prototype_pollution(obj[key], current_depth + 1, max_depth)
        elif isinstance(obj, list):
            for item in obj:
                GraphUtils.remove_prototype_pollution(item, current_depth + 1, max_depth)

    @staticmethod
    def remove_prefix(node_data, prefix):
        if node_data is None:
            return {}
        return {k[len(prefix):] if k.startswith(prefix) else k: v for k, v in node_data.items()}

    @staticmethod
    def convert_to_float(value):
        try:
            return float(value)
        except ValueError:
            return float('nan')

    @staticmethod
    def format_relative_err(value):
        """格式化相对误差为百分比，保留四位小数"""
        if value is None or math.isnan(value):
            return "NaN"
        else:
            return "{:.4%}".format(value)

    @staticmethod
    def nan_to_str(value):
        """将 NaN 转换为 'NaN' 字符串"""
        return "NaN" if math.isnan(value) else value

    @staticmethod
    def is_relative_to(path, base):
        abs_path = os.path.abspath(path)
        abs_base = os.path.abspath(base)
        return os.path.commonpath([abs_path, abs_base]) == str(abs_base)

    @staticmethod
    def bytes_to_human_readable(size_bytes, decimal_places=2):
        """
        将字节大小转换为更易读的格式（如 KB、MB、GB 等）。
        
        :param size_bytes: int 或 float，表示字节大小
        :param decimal_places: 保留的小数位数，默认为 2
        :return: str，人类可读的大小表示
        """
        if size_bytes == 0:
            return "0 B"

        units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
        unit_index = 0

        while size_bytes >= 1024 and unit_index < len(units) - 1:
            size_bytes /= 1024.0
            unit_index += 1

        return f"{size_bytes:.{decimal_places}f} {units[unit_index]}"

    @staticmethod
    def safe_save_data(data, run_name, tag):
        runs = GraphState.get_global_value('runs', {})
        run = runs.get(run_name) or run_name
        safe_base_dir = GraphState.get_global_value('logdir')
        if run is None or tag is None:
            error_message = 'The query parameters "run" and "tag" are required'
            return None, error_message
        try:
            # 构建文件路径并标准化
            file_path = os.path.join(run, tag)
            file_path = os.path.normpath(file_path)
            # 权限校验：检查目录是否有写权限
            if not os.access(run, os.W_OK):
                raise PermissionError(f"No write permission for directory: {run}\n")
            # 检查 run 目录是否存在，如果不存在则创建
            if not os.path.exists(run):
                os.makedirs(run, exist_ok=True)
                os.chmod(run, 0o750)
            # 检查 tag 是否为合法文件名
            if not re.match(FILE_NAME_REGEX, tag):
                raise ValueError(f"Invalid tag: {tag}.")
            # 检查文件路径是否合法，防止路径遍历攻击
            if not file_path.startswith(os.path.abspath(run)):
                raise ValueError(f"Invalid file path: {file_path}. Potential path traversal attack.\n")
            # 基础路径校验
            if not GraphUtils.is_relative_to(file_path, safe_base_dir):
                raise ValueError(f"Path out of bounds: {file_path}")
            if os.path.islink(file_path):
                raise RuntimeError("The target file is a symbolic link")
            if os.path.islink(run):
                raise RuntimeError(f"Parent directory contains a symbolic link")
            # # 尝试写入文件
            with open(file_path, "w", encoding="utf-8") as file:
                json.dump(data, file, ensure_ascii=False, indent=4)
            os.chmod(file_path, 0o640)
            # 最终校验（防御TOCTOU攻击）
            if os.path.islink(file_path):
                raise RuntimeError("The file has been replaced with a symbolic link")
            return True, None
        except (TypeError, ValueError) as e:
            logger.error(f"Invalid data: {e}")
            return None, 'Invalid data'
        except OSError as e:
            logger.error(f"Failed to create directory: {run}. Error: {e}\n")
            return None, 'failed to create directory {run}'
        except Exception as e:
            logger.error(f'Error: File "{file_path}" is not accessible. Error: {e}')
            return None, 'failed to save file'

    @staticmethod
    def safe_load_data(run_name, tag, only_check=False):
      
        runs = GraphState.get_global_value('runs', {})
        run_dir = runs.get(str(run_name)) or run_name
        """Load a single .vis file from a given directory based on the tag."""
        if run_dir is None or tag is None:
            error_message = 'The query parameters "run" and "tag" are required'
            return None, error_message
        try:
            file_path = os.path.join(run_dir, tag)
            file_path = os.path.normpath(file_path)  # 标准化路径
            # 解析真实路径（包含符号链接跟踪）
            real_path = os.path.realpath(file_path)
            safe_base_dir = GraphState.get_global_value('logdir')
            # 安全验证1：路径归属检查（防止越界访问）
            if not GraphUtils.is_relative_to(file_path, safe_base_dir):
                raise RuntimeError(f"Path out of bounds:")
            # 安全验证2：禁止符号链接文件
            if os.path.islink(file_path):
                raise RuntimeError(f"Detected symbolic link file")
            if os.path.islink(run_dir):
                raise RuntimeError(f"Parent directory contains a symbolic link")
            # 安全验证3：二次文件类型检查（防御TOCTOU攻击）
            if not os.path.isfile(real_path):
                raise RuntimeError(f"Path is not a regular file")
            # 安全检查4：文件存在性验证
            if not os.path.exists(real_path):
                raise FileNotFoundError(f"File does not exist")
            # 权限验证
            if not os.stat(real_path).st_mode & stat.S_IRUSR:
                raise PermissionError(f"File has no read permissions")
            # 文件大小验证
            if os.path.getsize(real_path) > MAX_FILE_SIZE:
                file_size = GraphUtils.bytes_to_human_readable(os.path.getsize(real_path))
                max_size = GraphUtils.bytes_to_human_readable(MAX_FILE_SIZE)
                raise RuntimeError(
                    f"File size exceeds limit ({file_size} > {max_size})")
            # 读取文件比较耗时，支持onlyCheck参数，仅进行安全校验
            if only_check:
                return True, None
            # 尝试解析 JSON 文件,校验文件内容是否合理
            with open(file_path, 'r', encoding='utf-8') as f:
                return json.load(f), None
        except json.JSONDecodeError:
            logger.error(f'Error: File "{file_path}" is not a valid JSON file!')
            return None, "File is not a valid JSON file!"
        except Exception as e:
            logger.error(f'Error: File "{file_path}" is not accessible. Error: {e}')
            return None, 'failed to load file'

    @staticmethod
    def find_config_files(run_name):
        """
        在指定目录下查找所有以 .vis.config 结尾的文件（不包括子目录）
        :param directory: 要搜索的目标目录路径
        :return: 包含所有匹配文件路径的列表
        """
        runs = GraphState.get_global_value('runs', {})
        run = runs.get(run_name)
        dir_path = Path(run)
        try:
            file_path = os.path.normpath(run)  # 标准化路径
            # 解析真实路径（包含符号链接跟踪）
            real_path = os.path.realpath(file_path)
            safe_base_dir = GraphState.get_global_value('logdir')
            # 安全验证1：路径归属检查（防止越界访问）
            if not GraphUtils.is_relative_to(file_path, safe_base_dir):
                raise RuntimeError(f"Path out of bounds:")
            # 安全验证2：禁止符号链接文件
            if os.path.islink(file_path):
                raise RuntimeError(f"Detected symbolic link file")
            # 安全检查3：文件存在性验证
            if not os.path.exists(real_path):
                raise FileNotFoundError(f"File does not exist")
            # 权限验证
            if not os.stat(real_path).st_mode & stat.S_IRUSR:
                raise PermissionError(f"File has no read permissions")
            if not dir_path.is_dir():
                raise ValueError(f"The provided path '{run}' is not a valid directory.")
            return [
                file.name for file in dir_path.iterdir()
                if file.is_file() and file.name.endswith('.vis.config')
            ]
        except Exception as e:
            logger.error(e)
            return []

    @staticmethod
    def compare_tag_names(a: str, b: str) -> int:
        """自然排序比较函数，支持路径格式和数字顺序"""

        # 辅助函数：标准化路径并分割组件
        def split_components(s: str) -> list:
            s = s.replace('\\', '/')  # 统一路径分隔符
            return [c for c in re.split(r'[/_]+', s) if c]  # 按 /_ 分割非空组件

        # 辅助函数：将组件拆分为数字/字符串混合的 tokens
        def tokenize(component: str) -> list:
            tokens = []
            buffer = []
            is_num = False

            for char in component:
                if char.isdigit() == is_num and buffer:
                    buffer.append(char)
                else:
                    if buffer:
                        tokens.append(int(''.join(buffer)) if is_num else ''.join(buffer))
                    buffer = [char]
                    is_num = char.isdigit()

            if buffer:
                tokens.append(int(''.join(buffer)) if is_num else ''.join(buffer))
            return tokens

        # 逐级比较组件
        a_comps = split_components(a)
        b_comps = split_components(b)

        for a_part, b_part in zip(a_comps, b_comps):
            a_tokens = tokenize(a_part)
            b_tokens = tokenize(b_part)

            # 比较 token 序列
            for a_tok, b_tok in zip(a_tokens, b_tokens):
                if isinstance(a_tok, int) and isinstance(b_tok, int):
                    if a_tok != b_tok:
                        return a_tok - b_tok
                elif isinstance(a_tok, int):
                    return -1  # 数字优先于字母
                elif isinstance(b_tok, int):
                    return 1
                elif a_tok != b_tok:
                    return -1 if a_tok < b_tok else 1

            # 处理子序列长度差异
            if len(a_tokens) != len(b_tokens):
                return len(a_tokens) - len(b_tokens)

        # 处理组件数量差异
        return len(a_comps) - len(b_comps)

    @staticmethod
    def sort_data(data: dict) -> dict:
        """自然排序比较函数，支持路径格式和数字顺序"""
        sorted_data = {}
        sorted_keys = sorted(data.keys(), key=cmp_to_key(GraphUtils.compare_tag_names))
        for k in sorted_keys:
            # 对每个键对应的值列表进行排序
            sorted_values = sorted(data[k], key=cmp_to_key(GraphUtils.compare_tag_names))
            sorted_data[k] = sorted_values

        return sorted_data

    @staticmethod
    def process_vis_file(dir_path, file, run_tag_pairs):
        file_path = os.path.join(dir_path, file)
        if os.path.isfile(file_path) and file.endswith('.vis'):
            run = dir_path
            run_name = os.path.basename(run)
            GraphState.set_global_value('runs', run, run_name)
            tag = file[:-4]  # Use the filename without extension as tag
            _, error = GraphUtils.safe_load_data(run_name, tag, True)
            if error:
                logger.error(f'Error: File run:"{run}, tag:{tag}" is not accessible. Error: {error}')
                return
            run_tag_pairs.setdefault(run_name, []).append(tag)
